@@ -1,4 +1,4 @@
-#define MAXIMUM_INITIAL_ID (1L << 31)
+#define MAXIMUM_INITIAL_ID 2600000000
 
 #define READ_NODE_CACHE_SIZE 10000
 #define READ_NODE_BLOCK_SHIFT 10l
