@@ -55,6 +55,7 @@ static uint32_t get_length(FILE *input)
   return ntohl(*((size_t *)buf));
 }
 
+#if 0
 static void *realloc_or_free(void *p, size_t len)
 {
   void *new = realloc(p, len);
@@ -65,6 +66,7 @@ static void *realloc_or_free(void *p, size_t len)
 
   return new;
 }
+#endif
 
 static BlockHeader *read_header(FILE *input, void *buf)
 {    
@@ -338,15 +340,13 @@ int processOsmDataDenseNodes(struct osmdata_t *osmdata, PrimitiveGroup *group, S
                 addIntItem(&(osmdata->tags), "osm_version", denseinfo->version[node_id], 0);
                 addIntItem(&(osmdata->tags), "osm_changeset", deltachangeset, 0);
                 
-#if 0
-                /* TODO */
                 if (deltauid != -1) { /* osmosis devs failed to read the specs */
-                    printuser(string_table->s[deltauser_sid]);
-                    printnumericattribute("osm_uid", deltauid);
+                    char * valstr;
+                    addIntItem(&(osmdata->tags), "osm_uid", deltauid, 0);
+                    valstr = calloc(string_table->s[deltauser_sid].len + 1, 1);
+                    memcpy(valstr, string_table->s[deltauser_sid].data, string_table->s[deltauser_sid].len);
+                    addItem(&(osmdata->tags), "osm_user", valstr,  0);
                 }
-                
-                printtimestamp("osm_timestamp", deltatimestamp);
-#endif
             }
             
             if (l < dense->n_keys_vals) {
